using System;
using System.Collections.Generic;
using System.Text;
using NHibernate;
using NHibernate.Metadata;
using NHibernate.Type;
using NHibernate.Cfg;
using System.Web;
using System.Configuration;
using System.Collections.Specialized;
using Common.Attributes;
using System.IO;
using NHibernate.Mapping.Attributes;
using System.Reflection;
using log4net;

namespace Common.Data.Hibernate
{
    public class NHibernateUtils
    {
        private static readonly ILog _log = LogManager.GetLogger(
            System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        /// <summary>
        /// Convenience method for the most common way of loading a hibernate configuration
        /// 
        /// <list>
        ///     <item>
        ///         If AppSettings contains an "nhibernate.config" setting, use it to load the file.  
        ///         (Relative references are allowed)
        ///     </item>
        ///     <item>
        ///         Otherwise fall back on NHibernate default Configure() method, which
        ///         looks in the AppConfig for NHibernate configuration info.
        ///     </item>
        ///     <item>
        ///         After configuring, calls <see cref="ProcessAttributeMappings()" />
        ///         to load any attribute-based mappings 
        ///     </item>
        /// </list>
        /// </summary>
        public static NHibernate.Cfg.Configuration LoadDefaultConfiguration()
        {
            NHibernate.Cfg.Configuration configuration = new NHibernate.Cfg.Configuration();
            String configPath = ConfigurationSettings.AppSettings["nhibernate.config"];
            if (configPath != null)
            {
                if (HttpContext.Current != null)
                {
                    configPath = HttpContext.Current.Request.MapPath(
                        configPath
                        );
                }
                else
                {
                    // relative path stuff for non-webapps?
                }

                configuration.Configure(configPath);
            }
            else
            {
                configuration.Configure();
            }
            ConfigureAttributeMappings(configuration);
            return configuration;
        }

		/// <summary>
		/// Searches for mapping attributes in each assembly and adds them to the configuration
		/// </summary>
		private static void ConfigureAttributeMappings(NHibernate.Cfg.Configuration configuration)
		{
			string attributeAssemblies = ConfigurationSettings.AppSettings["nhibernate.mapping.attributes.assemblies"];
			if (!string.IsNullOrEmpty(attributeAssemblies)) {
				string[] attrAsms = attributeAssemblies.Split(',');
				foreach (string asmName in attrAsms)
				{
					ProcessAttributeMappings(configuration, Assembly.LoadWithPartialName(asmName));
				}
			}
		}

        /// <summary>
        /// Searches for mapping attributes in each assembly and adds them to the configuration.
        /// 
        /// In order for the assembly to be loaded, it must have the <see cref="NHibernateAttributes" />
        /// attribute on the assembly.
        /// </summary>
		private static void ProcessAttributeMappings(NHibernate.Cfg.Configuration configuration, Assembly asm)
		{
			_log.Info("Process mappings in assembly: " + asm.FullName);

			MemoryStream stream = new MemoryStream();
			HbmSerializer.Default.Validate = true;

			// This is to provide a way to set global hibernate mapping attributes for all classes in the assembly
			object[] hmas = asm.GetCustomAttributes(typeof(Common.Attributes.HibernateMappingAttribute), false);
			if (hmas.Length > 0)
			{
				Common.Attributes.HibernateMappingAttribute hma =
					(Common.Attributes.HibernateMappingAttribute)hmas[0];

				if (hma.Assembly != null) HbmSerializer.Default.HbmAssembly = hma.Assembly;
				if (hma.AutoImportSpecified) HbmSerializer.Default.HbmAutoImport = hma.AutoImport;
				if (hma.DefaultAccess != null) HbmSerializer.Default.HbmDefaultAccess = hma.DefaultAccess;
				if (hma.DefaultCascade > 0) HbmSerializer.Default.HbmDefaultCascade = hma.DefaultCascade;
				if (hma.DefaultLazySpecified) HbmSerializer.Default.HbmDefaultLazy = hma.DefaultLazy;
				if (hma.Namespace != null) HbmSerializer.Default.HbmNamespace = hma.Namespace;
				if (hma.Schema != null) HbmSerializer.Default.HbmSchema = hma.Schema;
			}

			HbmSerializer.Default.Serialize(stream, asm);
			stream.Position = 0; // Rewind
			try
			{
				configuration.AddInputStream(stream);
			}
			catch (Exception ex)
			{
				throw new Exception(new UTF8Encoding().GetString(stream.ToArray()), ex);
			}
			stream.Close();
		}        

        public static IType GetPropertyType(
            ISessionFactory sessionFactory,
			Type classType,
            string propertyName)
        {
            string[] properties = propertyName.Split(new char[]{'.'});
            IClassMetadata currentMetadata = sessionFactory.GetClassMetadata(classType);
            IType currentType = null;
            foreach (string s in properties)
            {
                // get metadata from type, if necessary
                if (currentMetadata == null)
                {
                    currentMetadata = sessionFactory.GetClassMetadata(currentType.ReturnedClass);
                }
                // get type from metadata
                if ("id".Equals(s))
                {
                    currentType = currentMetadata.IdentifierType;
                }
                else
                {
                    currentType = currentMetadata.GetPropertyType(s);
                }
                
                currentMetadata = null;
            }
            return currentType;
        }

		public static IType GetType(TypeCode typeCode) 
        {
			switch (typeCode) {
				case TypeCode.Boolean:
					return NHibernateUtil.Boolean;
				case TypeCode.Byte:
					return NHibernateUtil.Byte;
				case TypeCode.Char:
					return NHibernateUtil.Character;
				case TypeCode.DateTime:
					return NHibernateUtil.DateTime;
				case TypeCode.Decimal:
					return NHibernateUtil.Decimal;
				case TypeCode.Double:
					return NHibernateUtil.Double;
				case TypeCode.Int16:
					return NHibernateUtil.Int16;
				case TypeCode.Int32:
					return NHibernateUtil.Int32;
				case TypeCode.Int64:
					return NHibernateUtil.Int64;
				case TypeCode.SByte:
					return NHibernateUtil.SByte;
				case TypeCode.Single:
					return NHibernateUtil.Single;
				case TypeCode.String:
					return NHibernateUtil.String;
				case TypeCode.UInt16:
					return NHibernateUtil.UInt16;
				case TypeCode.UInt32:
					return NHibernateUtil.UInt32;
				case TypeCode.UInt64:
					return NHibernateUtil.UInt64;
				default:
					return null;
			}
		}

        public static void CreateDatabase(NHibernate.Cfg.Configuration configuration)
        {
            NHibernate.Tool.hbm2ddl.SchemaExport se = 
                new NHibernate.Tool.hbm2ddl.SchemaExport(configuration);
            se.Create(true, true);
        }

        public static void DoInTransaction(ISession session, System.Action<ISession> action)
        {
			if (session.Transaction != null && session.Transaction.IsActive)
			{
				action(session);
			}
			else
			{
				ITransaction transaction = session.BeginTransaction();
				try
				{
					action(session);
					transaction.Commit();
				}
				catch (Exception)
				{
					transaction.Rollback();
					throw;
				}
			}
        }
	}
}
