using System;
using System.Collections.Generic;
using System.Text;
using NHibernate;
using System.Threading;
using System.Web;
using NHibernate.Cfg;

namespace Common.Data.Hibernate
{
    /// <summary>
    ///     Provides access to a single NHibernate session on a per request basis.
    ///     For web applications, the session is stored in HttpContext.Current.Items    
    ///     For Client applications, the session is stored in a named slot in the thread.
    /// 
    ///     In both cases the key is <see cref="SESSION_KEY"/>.
    /// </summary>
    public class SessionHolder
    {
        private static ISessionFactory _sessionFactory;

        /// <summary>
        /// Gets the current SessionFactory.  
        /// If it is not available, calls <see cref="Configure()"/>
        /// </summary>
        public static ISessionFactory SessionFactory
        {
            get {
                if (_sessionFactory == null) Configure();
                return _sessionFactory; 
            }
            set
            {
                _sessionFactory = value;
            }
        }

        private static IInterceptor _interceptor;

        public static IInterceptor Interceptor
        {
            get { return _interceptor; }
            set { _interceptor = value; }
        }


        /// <summary>
        /// Default configuration method, calls <see cref="NHibernateUtils.LoadDefaultConfiguration()"/>.
        /// </summary>
        private static void Configure()
        {
            Configure(
                NHibernateUtils.LoadDefaultConfiguration()
                );
        }

        /// <summary>
        /// Option that allows the SessionHolder to be passed a Configuration to use.
        /// 
        /// Only useful if the developer does not want to rely on 
        /// <see cref="NHibernateUtils.LoadDefaultConfiguration()"/>
        /// </summary>        
        public static void Configure(Configuration configuration)
        {
            _sessionFactory = configuration.BuildSessionFactory();
        }

        /// <summary>
        /// Key that is used for storing the session in its context
        /// </summary>
        public const string SESSION_KEY = "HibernateSession";

        /// <summary>
        /// Returns a thread-local session.
        /// 
        /// Does not auto-create the session.  See <see cref="Current"/> for that.
        /// </summary>
        protected static ISession ClientSession
        {
            get {
                ISession session = (ISession) Thread.GetData(Thread.GetNamedDataSlot(SESSION_KEY));
                return session;
            }
            set
            {
                Thread.SetData(Thread.GetNamedDataSlot(SESSION_KEY), value);
            }
        }


        /// <summary>
        /// Returns a session from HttpContext.Current.Items
        /// 
        /// Does nto auto-create the session
        /// </summary>
        protected static ISession WebSession
        {
            get 
            {
                return (ISession)HttpContext.Current.Items[SESSION_KEY];
            }
            set
            {
                HttpContext.Current.Items[SESSION_KEY] = value;
            }
            
        }

        /// <summary>
        /// Convenience property - just calls <see cref="CreateTemplate()"/>
        /// </summary>
        public static NHibernateTemplate Template
        {
            get
            {
                return CreateTemplate();
            }
        }

        /// <summary>
        /// Convenience property - constructs a new template around the session returned by
        /// <see cref="Current"/>
        /// </summary>
        /// <returns></returns>
        public static NHibernateTemplate CreateTemplate()
        {
            return new NHibernateTemplate(Current);
        }

        /// <summary>
        /// Convenience property, returns a new generic <see cref="NHibernateTemplate<T>"/>
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public static NHibernateTemplate<T> CreateTemplate<T>()
        {
            return new NHibernateTemplate<T>(Current);
        }
        
        /// <summary>
        ///Convenience property, although this is the most commonly used property in this class.
        /// 
        /// Simply calls <see cref="GetCurrentSession(true)"/>
        /// </summary>
        public static ISession Current
        {
            get 
            {
                return GetCurrentSession(true);
            }
            private set
            {
                if (HttpContext.Current != null)
                {
                    WebSession = value;
                }
                else
                {
                    ClientSession = value;
                }
            }
        }

        public static ISession OpenSession()
        {
            if (Interceptor != null) return SessionFactory.OpenSession(Interceptor);
            return SessionFactory.OpenSession();
        }

        /// Returns the current session given the context, automatically creating it
        /// if <param name="autoCreate"/> is true.  
        /// 
        /// If HttpContext.Current is available, <see cref="WebSession"/>
        /// is used.  Otherwise <see cref="ClientSession"/> is used.
        public static ISession GetCurrentSession(bool autoCreate)
        {
            ISession session = null;
            if (HttpContext.Current != null)
            {
                session = WebSession;
                if (session == null && autoCreate)
                {
                    session = OpenSession();
                    WebSession = session;
                }
            }
            else
            {
                session = ClientSession;
                if (session == null && autoCreate)
                {
                    session = OpenSession();
                    ClientSession = session;
                }
            }
            return session;
        }

        /// <summary>
        /// Checks whether there is a current session open
        /// </summary>
        public bool IsCurrentSessionOpen
        {
            get
            {
                ISession session = GetCurrentSession(false);
                if (session != null && session.IsOpen) return true;
                return false;
            }
        }

        /// <summary>
        /// If there is a current session, close it and remove it from the context.
        /// </summary>
        public static void CloseCurrentSession()
        {
            ISession current = GetCurrentSession(false);
            if (current != null)
            {
                current.Close();
                Current = null;
            }
        }

    }
}
