using System;
using System.Collections;
using System.Text;
using System.Data;
using System.Reflection;
using System.Collections.ObjectModel;
using System.Collections.Generic;
using SC = System.Collections;
using Spring.Objects;
using System.ComponentModel;
using System.Collections.Specialized;
using System.Text.RegularExpressions;
using Spring.Expressions;

namespace Common.Data
{
    public class ObjectDataTableAdapter
    {
        public ObjectDataTableAdapter() { }
        
        public ObjectDataTableAdapter(SC.IList data)
        {
            _data = data;
        }

        private Collection<PropertyColumn> _columns = new Collection<PropertyColumn>();

        public Collection<PropertyColumn> Columns
        {
            get { return _columns; }
        }

        private IList _data;

        public IList Data
        {
            get { return _data; }
            set { 
                _data = value;
                _initialized = false;
            }
        }


        /// <summary>
        /// This is a special property name that will access the item 
        /// in the underlying collection, rather than one of its properties
        /// 
        /// This is used when the backing collection is collection of primitives
        /// </summary>
        public const string ValueKey = "_Value_";

        private List<PropertyColumn> _internalMappings;

        private bool _autoGenerateColumns = true;

        [DefaultValue(true)]
        public bool AutoGenerateColumns
        {
            get { return _autoGenerateColumns; }
            set { _autoGenerateColumns = value; }
        }

        private StringCollection _propertiesToIgnore = new StringCollection();
        public StringCollection PropertiesToIgnore
        {
            get { return _propertiesToIgnore; }
        }

        bool _initialized = false;

        private string _anonymousColumnPrefix = "Column";

        /// <summary>
        /// When auto-generating columns that are not property-based.
        /// 
        /// For example, the collection contains primitives or an array.
        /// In this case, there is no property name to use, so an 
        /// "anonymous" column is generated.
        /// 
        /// To specify column names for array indices and primitives,
        /// disable AutoGenerateColumns and instead create PropertyColumn
        /// entries, setting the PropertyName to the index (i.e. "[0]")
        /// or ValueKey for the primitive value
        /// </summary>
        [DefaultValue("Column")]
        public string AnonymousColumnPrefix
        {
            get { return _anonymousColumnPrefix; }
            set { _anonymousColumnPrefix = value; }
        }


        private void EnsureInitialized()
        {
            if (_initialized) return;
            _internalMappings  = new List<PropertyColumn>();
            if (AutoGenerateColumns)
            {
                Type type = GetDataObjectType();
                // if type is null, we cannot generate columns
                // type is only null if there are no results
                if (type != null)
                {
                    if (type.IsPrimitive || type == typeof(string))
                    {
                        _internalMappings.Add(new PropertyColumn(ValueKey, AnonymousColumnPrefix + "0", type));
                    }
                    else if (type.IsArray && _data.Count > 0)
                    {
                        int colCount = ((IList)_data[0]).Count;

                        for (int i = 0; i < colCount; i++)
                        {
                            _internalMappings.Add(
                                new PropertyColumn(
                                    "[" + i + "]",
                                    AnonymousColumnPrefix + i
                                ));
                        }
                    }
                    else
                    {
                        PropertyInfo[] pi = type.GetProperties();

                        foreach (PropertyInfo p in pi)
                        {
                            _internalMappings.Add(new PropertyColumn(p.Name, p.Name, p.PropertyType));
                        }
                    }
                }
                
            }
            _internalMappings.RemoveAll(
                delegate(PropertyColumn mapping)
                {
                    return PropertiesToIgnore.Contains(mapping.ColumnName);
                }
            );

            _internalMappings.AddRange(_columns);
            _initialized = true;            
        }

        private Type GetDataObjectType()
        {
            if (Type != null) return Type;
            return GuessType();
        }

        private Type GuessType()
        {
            Type type = null;

            foreach (object o in _data)
            {
                if (o == null) continue;
                Type tO = o.GetType();
                if (type == null) type = tO;
                else
                {
                    if (tO.IsAssignableFrom(type))
                        type = tO;
                }
            }
            return type;
        }

        public PropertyColumn GetMappedProperty(string name)
        {
            EnsureInitialized();
            foreach (PropertyColumn p in _internalMappings)
            {
                if (p.ColumnName == name) return p;
            }
            return null;
        }

        public object GetProperty(object obj, string name)
        {
            EnsureInitialized();
            ObjectWrapper wrapper = new ObjectWrapper(obj);
            return GetProperty(wrapper, GetMappedProperty(name));
        }

        
        public object GetProperty(IObjectWrapper wrapper, PropertyColumn mapping)
        {            
            string propertyName = mapping.PropertyName;
            if (propertyName == ValueKey)
                return wrapper.WrappedInstance;
            try
            {
                // special handling for raw indexed properties
                if (propertyName.StartsWith("["))
                {
                    return ExpressionEvaluator.GetValue(wrapper.WrappedInstance, propertyName);
                }
                return wrapper.GetPropertyValue(propertyName);
            }
            catch (NullValueInNestedPathException)
            {
                // always ignore nulls in path
            }
            catch (Exception)
            {
                if (!IgnoreAllReflectionExceptions) 
                    throw;                
            }
            return null;
        }

        private bool _ignoreAllReflectionExceptions = true;

        [DefaultValue(true)]
        public virtual bool IgnoreAllReflectionExceptions
        {
            get { return _ignoreAllReflectionExceptions; }
            set { _ignoreAllReflectionExceptions = value; }
        }


        public bool IsNullable(Type type)
        {
            return
                type.IsGenericType &&
                type.GetGenericTypeDefinition() == typeof(Nullable<>);
        }

        public void SetProperty(object obj, string name, object value)
        {
            EnsureInitialized();
            ObjectWrapper wrapper = new ObjectWrapper(obj); 
            SetProperty(wrapper, GetMappedProperty(name), value);
        }

        /// <summary>
        /// This will update the backing object with the property value,
        /// managing null conversions if necessary.
        /// </summary>
        private void SetProperty(IObjectWrapper wrapper, PropertyColumn mapping, object value)
        {            
            string propertyName = mapping.PropertyName;
            Type type = wrapper.GetPropertyType(propertyName);
        
            if (IsNullable(type) && value == System.DBNull.Value)
                value = null;

            PropertyValue v = new PropertyValue(propertyName, value);
            try
            {
                if (propertyName.StartsWith("["))
                {
                    ExpressionEvaluator.SetValue(wrapper.WrappedInstance, propertyName, value);
                    return;
                }
                wrapper.SetPropertyValue(v);
            }
            catch (Exception) { }
                        
        }

        private Type GetDataTableStorageType(Type type)
        {            
            if (IsNullable(type))
                return Nullable.GetUnderlyingType(type);
            return type;
        }

        public DataTable GenerateDataTable()
        {
            EnsureInitialized();
            DataTable dt = new DataTable();

            // Generate schema
            foreach (PropertyColumn p in _internalMappings)
            {      
                // if Type was not specified, this assumes it's a string
                if (p.Type == null) dt.Columns.Add(p.ColumnName);
                else
                {
                    Type type = GetDataTableStorageType(p.Type);
                    dt.Columns.Add(p.ColumnName, type);
                }
            }

            // Add data
            foreach (object obj in _data)
            {
                IObjectWrapper wrapper;
                
                if (obj != null)
                    wrapper = new ObjectWrapper(obj);
                else 
                    wrapper =  new ObjectWrapper();
                
                object[] row = new object[_internalMappings.Count];
                

                int i = 0;

                foreach (PropertyColumn p in _internalMappings)
                {
                    object value = GetProperty(wrapper, p);
                    // lazily set column type = value type, if still null
                    if (p.Type == null && value!=null)
                    {
                        dt.Columns[i].DataType = GetDataTableStorageType(value.GetType());
                    }
                        
                    row[i++] = value;

                }
                dt.Rows.Add(row);

            }
            // TODO: leaving this out as it is unnecessary and barely tested
            // but it seems like it would be useful in some situations.
            // dt.ColumnChanging += new DataColumnChangeEventHandler(dt_ColumnChanging);
            return dt;
        }

        /// <summary>
        /// When a cell in the datatable is modified, this event
        /// will update the backing property.
        /// 
        /// It will not refresh any other values in the datatable
        /// </summary>
        void dt_ColumnChanging(object sender, DataColumnChangeEventArgs e)
        {
            int index = e.Row.Table.Rows.IndexOf(e.Row);
            PropertyColumn pm = GetMappedProperty(e.Column.ColumnName);            
            object obj = _data[index];

            if (pm.ColumnName == ValueKey)
            {
                _data[index] = e.ProposedValue;
                return;
            }

            SetProperty(obj, e.Column.ColumnName, e.ProposedValue);
        }

        private Type _type;

        public virtual Type Type
        {
            get { return _type; }
            set { _type = value; }
        }


    }

    public class PropertyColumn
    {
        public PropertyColumn() { }

        public PropertyColumn(string propertyName, string columnName, Type type) {
            _sourceProperty = propertyName;
            _columnName = columnName;
            _type = type;
        }

        public PropertyColumn(string propertyName, string columnName) :this(propertyName,columnName,null) { }

        public PropertyColumn(string propertyName) : this(propertyName, propertyName) { }

        private string _sourceProperty;

        public string PropertyName
        {
            get { return _sourceProperty; }
            set { _sourceProperty = value; }
        }

        private string _columnName;

        public string ColumnName
        {
            get { return _columnName; }
            set { _columnName = value; }
        }

        private Type _type;

        public Type Type
        {
            get { return _type; }
            set { _type = value; }
        }


    }
}
