using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing.Design;
using System.Configuration;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;
using Common.Data.Hibernate;
using System.Collections.ObjectModel;
using NHibernate;

namespace Common.Web.CustomControls
{

    /// <summary>
    /// Originally inspired by a control from Paul Wilson.  He wished the comment
    /// below to be included
    /// 
    /// // This control was adapted from a sample as described below
    /// //**************************************************************//
    /// // Paul Wilson -- www.WilsonDotNet.com -- Paul@WilsonDotNet.com //
    /// // Feel free to use and modify -- just leave these credit lines //
    /// // I also always appreciate any other public credit you provide //
    /// //**************************************************************//
    /// </summary>
    [
    PersistChildren(false), 
    ParseChildren(true),     
    ]
    public class NHibernateDataSource : System.Web.UI.DataSourceControl
    {
        private string typeName;
		
        static private readonly string[] viewNames = { "DefaultView" };

        public NHibernateDataSource()
        {
            this.dataView = new NHibernateDataView(this, viewNames[0]);
        }

        #region Properties

        private NHibernateDataView dataView;

        /// <summary>
        /// Convenience property to get this data source's data view
        /// </summary>
        [Browsable(false)]
        public NHibernateDataView DataView
        {
            get { return dataView; }
        }

        private string _query;

        [Category("Data"),
                 DefaultValue((string)null),
                 Description("The HQL query (without ORDER BY -- use DefaultSortExpression)")]
        public string Query
        {
            get { return this._query; }
            set
            {
                if (!value.Equals(this._query))
                {
                    this._query = value;
                    this.RaiseDataSourceChangedEvent(EventArgs.Empty);
                }
            }
        }

        [Category("Data"),
         DefaultValue(""),
         Description("The NHibernate entity Type to be queried and bound to")]
        public string TypeName
        {
            get { return this.typeName; }
            set
            {
                if (!value.Equals(this.typeName))
                {
                    this.typeName = value;
                    this.RaiseDataSourceChangedEvent(EventArgs.Empty);
                }
            }
        }
        
        private bool _excludeEmptySelectParameters;

        [DefaultValue(false), 
        Category("Behavior"),
        Description("Do not generate Where clause fragments for parameters with empty values")
        ]
        public bool ExcludeEmptySelectParameters
        {
            get { return _excludeEmptySelectParameters; }
            set { _excludeEmptySelectParameters = value; }
        }

        private Type dataObjectType;
        [Browsable(false), Description("The Type corresponding to TypeName")]
        public Type DataObjectType
        {
            get
            {
                if (dataObjectType == null && !string.IsNullOrEmpty(TypeName))
                {
                    dataObjectType = System.Web.Compilation.BuildManager.GetType(TypeName, true, false);
                }
                return dataObjectType;
            }
            set
            {
                dataObjectType = value;                
            }
        }

        private void EnsureTypeInfo()
        {
            if (_identifierProperty != null) return;
            if (DataObjectType == null) return;

            NHibernate.Metadata.IClassMetadata md = 
                Session.SessionFactory.GetClassMetadata(DataObjectType);
            _identifierType = md.IdentifierType.ReturnedClass;
            _identifierProperty = md.IdentifierPropertyName;

        }

        private Type _identifierType;
        /// <summary>
        /// Used when loading or deleting an entity
        /// </summary>
        [Category("Data"), Browsable(false), Description("The Type of the entity's identifier")]
        internal Type IdentifierType
        {
            get {
                EnsureTypeInfo();
                return _identifierType; 
            }
        }

        private string _identifierProperty;
        [Category("Data"), Browsable(false)]
        internal string IdentifierProperty
        {
            get {
                EnsureTypeInfo();
                return _identifierProperty; 
            }
        }

        [Category("Data"), 
        PersistenceMode(PersistenceMode.InnerProperty), 
        DefaultValue((string)null), 
        Editor("System.Web.UI.Design.WebControls.ParameterCollectionEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor)), MergableProperty(false)]
        public ParameterCollection SelectParameters
        {
            get { return this.dataView.SelectParameters; }
        }


        private bool _enablePaging;

        /// <summary>
        /// NHibernate can page, but if it pages, there is no
        /// way to count the results of the query.  
        /// 
        /// In order to count the results, then, a separate 
        /// "select count" query must be issued.  See 
        /// <see cref="AutoGenerateCountQuery"/>.
        /// 
        /// Note that the ASP.NET databound controls are smart
        /// enough to perform pagination on their own.  
        /// 
        /// So you can still get pagination on GridView without enabling
        /// pagination on the data source.  It will simply return all
        /// the results of the query and GridView will slice and
        /// display only the results for that page.
        /// 
        /// Depending on data size, this may also be more performant 
        /// than limiting the data returned by NHibernate but and issuing 
        /// a subsequent count query.
        /// 
        /// Therefore, this property defaults to being false.
        /// </summary>
        [Category("Behavior"), DefaultValue(false),
         Description(
             "Enable NHibernate paging, which will disable total count " +
             "retrieval unless AutoGenerateCountQuery is enabled.")
        ]
        public virtual bool EnablePaging
        {
            get { return _enablePaging; }
            set { _enablePaging = value; }
        }

        private Collection<QueryFragment> _whereFragments = new Collection<QueryFragment>();

        [
        Category("Data"),
        Description("QueryFragments which may be added to the WHERE clause at runtime"),
        PersistenceMode(PersistenceMode.InnerProperty),
        DefaultValue((string[])null)
        ]
        public virtual Collection<QueryFragment> WhereFragments
        {
            get { return _whereFragments; }
        }

        private bool _AutoGenerateWhereClauseFromSelectParams = true;

        [Category("Behavior"), DefaultValue(true),
        Description("Generate a WHERE clause fragment for each SelectParameter")]
        public virtual bool AutoGenerateWhereClauseFromSelectParams
        {
            get { return _AutoGenerateWhereClauseFromSelectParams; }
            set { _AutoGenerateWhereClauseFromSelectParams = value; }
        }

        private string _defaultOrderBy;

        [Category("Behavior"),
        Description("The sort expression that will be used if none are specified in the select arguments")]
        public virtual string DefaultSortExpression
        {
            get { return _defaultOrderBy; }
            set { _defaultOrderBy = value; }
        }

        private string _wildCardCharacter = "*";

        [DefaultValue("*"),
        Description("The character that will be replaced with the corresponding database wildcard (usually \"%\")")]
        public virtual string WildCard
        {
            get { return _wildCardCharacter; }
            set { _wildCardCharacter = value; }
        }

        private ParameterExtenderCollection _parameterExtenders = new ParameterExtenderCollection();
        [
        Category("Data"),
        Description("Extra configuration about how SelectParameters should be handled"),
        PersistenceMode(PersistenceMode.InnerProperty),
        DefaultValue((string[])null)
        ]
        public ParameterExtenderCollection ParameterExtenders
        {
            get { return _parameterExtenders; }
        }

        private bool _generateDataTable = false;

        [Category("Behavior"), DefaultValue(false),
        Description("Put the returned result set into a DataTable")       
        ]
        public bool GenerateDataTable
        {
            get { return _generateDataTable; }
            set { _generateDataTable = value; }
        }

        private DataTableAdapterConfig _dataTableConfig = new DataTableAdapterConfig();

        [
        Category("Behavior"),
        Description("Extra configuration about how to generate a DataTable from the results"),
        PersistenceMode(PersistenceMode.InnerProperty),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Visible),
        TypeConverter(typeof(ExpandableObjectConverter)),
        DefaultValue((string)null)
        ]
        public DataTableAdapterConfig DataTableConfig
        {
            get { return _dataTableConfig; }
        }

        private bool _autoGenerateCountQuery = true;

        /// <summary>
        /// When paging is enabled, there is no way to know how 
        /// many TOTAL results there are without executing a separate 
        /// count query.
        /// 
        /// If AutoGenerateCountQuery is enabled, then when paging is
        /// used, the data source will use simple parsing to replace 
        /// the SELECT clause of the HQL query with "select count(*)"
        /// and then execute the query in order to get a count.
        /// 
        /// NOTE: This will work for simple queries, but not for all queries
        /// 
        /// For example, the query:
        /// 
        /// <example>select elements(cat.Kittens) from Eg.Cat cat</example>
        /// 
        /// Would generate a count query of
        /// 
        /// <example>select Count(*) from Eg.Cat cat</example>
        /// </summary>
        [Category("Behavior"), DefaultValue(true),
        Description("Used with EnablePaging, generate a Select Count(*) query from the original query")]
        public virtual bool AutoGenerateCountQuery
        {
            get { return _autoGenerateCountQuery; }
            set { _autoGenerateCountQuery = value; }
        }

        private ISession _session;
        /// <summary>
        /// The ISession used by this datasource
        /// 
        /// By default this will be automatically created
        /// with a call to <see cref="CreateSession()"/>.
        /// 
        /// However, it may also be set programatically.
        /// </summary>
        [Browsable(false)]
        public virtual ISession Session
        {
            get
            {
                if (_session == null || !_session.IsOpen)
                {
                    _session = CreateSession();
                }
                return _session;
            }
            set
            {
                _session = value;
            }
        }

        private bool _autoCommit = true;
        /// <summary>
        /// If true, Session.Transaction will be committed
        /// whenever ExecuteUpdate, ExecuteInsert, or ExecuteDelete
        /// are called.
        /// 
        /// Otherwise, the developer may programmatically commit the 
        /// transaction by calling <see cref="NHibernateDataView.CommitAndHandleRollback()"/> 
        /// or access the transaction directly via <see cref="Session"/>.Transaction.
        /// 
        /// Note: I've found that disabling AutoCommit usually doesn't do much
        /// unless FlushMode is also set to something Commit or Never
        /// </summary>
        [DefaultValue(true),
        Category("Behavior"),
        Description("Automatically commit any updates, inserts, or deletions")]
        public bool AutoCommit
        {
            get { return _autoCommit; }
            set { _autoCommit = value; }
        }

        private FlushMode _flushMode = FlushMode.Commit;
        /// <summary>
        /// The flush mode to apply to all newly created Sessions
        /// 
        /// Defaults to FlushMode.Commit because of the potential for multiple
        /// datasources on a page to interfere with each other.
        /// </summary>
        [DefaultValue(FlushMode.Commit),
        Category("Behavior"),
        Description("FlushMode to use with all newly created sessions")]
        public FlushMode FlushMode
        {
            get { return _flushMode; }
            set { _flushMode = value; }
        }



        #endregion

        #region Events

        /// <summary>Occurs when a <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Delete"></see> operation has completed.</summary>
        public event EventHandler Deleted;
        internal virtual void OnDeleted(EventArgs e)
        {
            if (Deleted != null) Deleted(this, e);
        }

        /// <summary>Occurs before a <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Delete"></see> operation.</summary>
        public event EventHandler<NHibernateEntityCancelEventArgs> Deleting;
        internal virtual void OnDeleting(NHibernateEntityCancelEventArgs e)
        {
            if (Deleting != null) Deleting(this, e);
        }

        /// <summary>Occurs when an <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Insert"></see> operation has completed.</summary>
        public event EventHandler<NHibernateEntityEventArgs> Inserted;
        internal virtual void OnInserted(NHibernateEntityEventArgs e)
        {
            if (Inserted != null) Inserted(this, e);
        }
        /// <summary>Occurs before an <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Insert"></see> operation.</summary>
        public event EventHandler<NHibernateEntityCancelEventArgs> Inserting;
        internal virtual void OnInserting(NHibernateEntityCancelEventArgs e)
        {
            if (Inserting != null) Inserting(this, e);
        }

        /// <summary>Occurs after the <see cref="T:System.Web.UI.WebControls.ObjectDataSourceView"></see> object creates an instance of the type that is identified by the <see cref="P:System.Web.UI.WebControls.ObjectDataSourceView.TypeName"></see> property.</summary>
        public event EventHandler<NHibernateEntityEventArgs> ObjectCreated;
        internal virtual void OnObjectCreated(NHibernateEntityEventArgs e)
        {
            if (ObjectCreated != null) ObjectCreated(this, e);
        }

        /// <summary>Occurs before the <see cref="T:System.Web.UI.WebControls.ObjectDataSourceView"></see> object creates an instance of the type that is identified by the <see cref="P:System.Web.UI.WebControls.ObjectDataSourceView.TypeName"></see> property.</summary>
        public event EventHandler<NHibernateEntityEventArgs> ObjectCreating;
        internal virtual void OnObjectCreating(NHibernateEntityEventArgs e)
        {
            if (ObjectCreating != null) ObjectCreating(this, e);
        }

        /// <summary>Occurs when a data retrieval operation has completed.</summary>
        public event EventHandler Selected;
        internal virtual void OnSelected(EventArgs e)
        {
            if (Selected != null) Selected(this, e);
        }

        /// <summary>Occurs before a data retrieval operation.</summary>
        public event EventHandler<NHibernateSelectingArgs> Selecting;
        internal virtual void OnSelecting(NHibernateSelectingArgs e)
        {
            if (Selecting != null) Selecting(this, e);
        }

        /// <summary>Occurs when an <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Update"></see> operation has completed.</summary>
        public event EventHandler Updated;
        internal virtual void OnUpdated(EventArgs e)
        {
            if (Updated != null) Updated(this, e);
        }

        /// <summary>Occurs before an <see cref="Overload:System.Web.UI.WebControls.ObjectDataSourceView.Update"></see> operation.</summary>
        public event EventHandler<NHibernateEntityCancelEventArgs> Updating;
        internal virtual void OnUpdating(NHibernateEntityCancelEventArgs e)
        {
            if (Updating != null) Updating(this, e);
        }

        public event EventHandler<NHibernateEntityLoadingArgs> LoadingEntity;
        internal virtual void OnLoadingEntity(NHibernateEntityLoadingArgs e)
        {
            if (LoadingEntity != null) LoadingEntity(this,e);
        }

        public event EventHandler CreatingSession;
        internal virtual void OnCreatingSession(EventArgs e)
        {
            if (CreatingSession != null) CreatingSession(this, e);
        }

        #endregion

        #region Extension Points

        /// <summary>
        /// This method is called in order to first get a 
        /// valid NHibernate ISession to work with.
        /// 
        /// The default implementation just calls 
        /// <see cref="SessionHolder.Current"/>.
        /// 
        /// This will also set FlushMode on the session
        /// </summary>
        protected virtual ISession CreateSession()
        {
            _session = null;
            
            // Allow creation of session via event
            OnCreatingSession(EventArgs.Empty);
            if (_session != null) 
                return _session;

            ISession session = SessionHolder.Current;
            if (FlushMode != FlushMode.Unspecified)
                session.FlushMode = FlushMode;
            return session;
        }

        /// <summary>
        /// Create a new instance for an insert operation
        /// </summary>        
        public virtual object CreateNewInstance()
        {
            NHibernateEntityEventArgs args = new NHibernateEntityEventArgs(null);
            OnObjectCreating(args);
            object value = null;            
            if (args.Entity != null)
                value = args.Entity;
            else
                value = Activator.CreateInstance(DataObjectType);
            OnObjectCreated(args);
            return value;
        }

        #endregion

        protected override DataSourceView GetView(string viewName)
        {
            return this.dataView;
        }

        protected override System.Collections.ICollection GetViewNames()
        {
            return viewNames;
        }

    }
}