using System;
using System.Text;
using System.Reflection;
using System.Data;
using System.Collections;
using System.Collections.Generic;
using System.Configuration;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;
using Common.Data.Hibernate;
using NHibernate;
using NHibernate.Metadata;
using Common.Util;
using NHibernate.Type;
using Common.Data;
using System.Web.Compilation;
using System.ComponentModel;


namespace Common.Web.CustomControls
{

    /// <summary>
    /// Originally inspired by a control from Paul Wilson.  He wished the comment
    /// below to be included
    /// 
    /// // This control was adapted from a sample as described below
    /// //**************************************************************//
    /// // Paul Wilson -- www.WilsonDotNet.com -- Paul@WilsonDotNet.com //
    /// // Feel free to use and modify -- just leave these credit lines //
    /// // I also always appreciate any other public credit you provide //
    /// //**************************************************************//
    /// </summary>
    public class NHibernateDataView : DataSourceView
    {

        public NHibernateDataView(NHibernateDataSource owner, string name)
            : base(owner, name)
        {
            this.owner = owner;
        }

        private NHibernateDataSource owner;

        
        #region Properties

        public override bool CanSort { get { return true; } }
        public override bool CanInsert { get { return true; } }
        public override bool CanUpdate { get { return true; } }
        public override bool CanDelete { get { return true; } }

        /// <summary>
        /// Paging can be enabled or disabled on the DataSource
        /// </summary>
        public override bool CanPage { get { return owner.EnablePaging; } }

        /// <summary>
        /// This is easily done if there is no paging.
        /// 
        /// If the datasource is asked to page the results, 
        /// then a second query must be issued in order to
        /// get the total row count.
        /// 
        /// <seealso cref="NHibernateDataSource.AutoGenerateCountQuery"/>
        /// <seealso cref="NHibernateDataSource.EnablePaging"/>
        /// </summary>
        public override bool CanRetrieveTotalRowCount
        {
            get
            {
                return !CanPage || owner.AutoGenerateCountQuery;
            }
        }

        private ParameterCollection _selectParameters;
        public ParameterCollection SelectParameters
        {
            get
            {
                if (this._selectParameters == null)
                {
                    this._selectParameters = new ParameterCollection();
                }
                return this._selectParameters;
            }
        }

        /// <summary>
        /// Private map of parameters and values
        /// TODO: replace this with a list of ParameterInfo?        
        /// </summary>
        private Dictionary<string, object> _innerParameters = new Dictionary<string, object>();

        /// <summary>
        /// Map of parameter values.  As the datasource processes declared parameters,
        /// their values are loaded into this map, which is also available for
        /// programmatic manipulation of parameters.
        /// </summary>
        [Browsable(false)]                
        public IDictionary<string, object> QueryParameters
        {
            get
            {
                return _innerParameters;
            }
        }

        /// <summary>
        /// The ISession used by this dataview
        /// </summary>
        protected virtual ISession Session
        {
            get
            {                
                return owner.Session;
            }            
        }

        private ObjectDataTableAdapter _dataTableAdapter;

        /// <summary>
        /// If <see cref="NHibernateDataSource.GenerateDataTable"/> is true,
        /// an <see cref="ObjectDataTableAdapter"/> will be used to wrap and
        /// generate the query results.
        /// 
        /// It will be available via this property after a Select or Update has
        /// been executed.
        /// </summary>
        public ObjectDataTableAdapter DataTableAdapter
        {
            get
            {
                if (_dataTableAdapter == null)
                    _dataTableAdapter = CreateDataTableAdapter();
                return _dataTableAdapter;
            }
        }


        private IList _resultList;
        /// <summary>
        /// The result list from the query.
        /// This will be available once a Select has been executed.
        /// </summary>
        public IList ResultList
        {
            get { return _resultList; }
        }

        private DataTable _resultDataTable;

        /// <summary>
        /// If <see cref="NHibernateDataSource.GenerateDataTable"/> is true,
        /// this convenience property will provide access to the generated
        /// DataTable after a Select has been called.
        /// </summary>
        public DataTable ResultDataTable
        {
            get { return _resultDataTable; }
        }

        #endregion

        #region Private Implementation

        /// <summary>
        /// Taking into account any ParameterExtenders,
        /// generate query fragments from SelectParameters
        /// </summary>        
        private List<QueryFragment> GenerateWhereFragmentsFromSelectParameters()
        {
            List<QueryFragment> list = new List<QueryFragment>();

            foreach (Parameter p in SelectParameters)
            {

                string propertyName = p.Name;
                string compareOperator = "=";
                ParameterExtender pe = owner.ParameterExtenders[p.Name];
                if (pe != null)
                {
                    if (!string.IsNullOrEmpty(pe.PropertyName))
                        propertyName = pe.PropertyName;
                    compareOperator = pe.WhereCompareOperator;
                }
                QueryFragment clause = new QueryFragment();
                clause.Expression = "o." + propertyName + " " + compareOperator + " :" + p.Name;
                List<string> dependencies = new List<string>();
                dependencies.Add(p.Name);
                if (pe!=null && pe.ParameterDependencies != null)
                {
                    dependencies.AddRange(pe.ParameterDependencies);
                }

                clause.ParameterDependencies = dependencies.ToArray();
                list.Add(clause);
            }

            return list;
        }


        /// <summary>
        /// Build a string where clause from the list of QueryFragments
        /// </summary>
        protected virtual string BuildWhereClause(List<QueryFragment> parms)
        {
			if (parms.Count == 0) return "";

            StringBuilder sb = new StringBuilder(" where ");
            
            string delimiter = " and ";
            foreach (QueryFragment entry in parms)
            {
                sb.Append(entry.Expression);
                sb.Append(delimiter);
            }
            sb.Remove(sb.Length - delimiter.Length, delimiter.Length);

            return sb.ToString();
        }

        private bool IsEmpty(object paramValue)
        {
            return paramValue == null || (
                paramValue is ValueType &&
                paramValue.Equals(Activator.CreateInstance(paramValue.GetType()))
                );
        }

        /// <summary>
        /// Remove any Where fragments whose dependencies are not met
        /// </summary>        
        protected virtual void FilterUnfulfilledDependencies(List<QueryFragment> whereClauses)
        {

            // find and remove fragments whose dependencies are missing
            whereClauses.RemoveAll(
                delegate(QueryFragment frag)
                {
                    foreach(string dependency in frag.ParameterDependencies){
                        bool negative = dependency.StartsWith("!");
                        if (negative)
                        {
                            string actualDependency = dependency.Substring(1);
                            if (!_innerParameters.ContainsKey(actualDependency)) continue;
                            if (!IsEmpty(_innerParameters[actualDependency])) return true;
                        }
                        else
                        {
                            if (!_innerParameters.ContainsKey(dependency)) return true;
                            if (IsEmpty(_innerParameters[dependency])) return true;
                        }                      
                    }
                    return false;
                });

        }

        /// <summary>
        /// Find any ControlParameterExtender instances and add their extension
        /// information to the DataSource's ParameterExtenders
        /// 
        /// Get values from SelectParameters, excluding empty values depending on
        /// <see cref="NHibernateDataSource.ExcludeEmptySelectParameters"/>
        /// 
        /// Also, add any prefixes or suffixes and perform wildcard replacement
        /// </summary>
        protected void ProcessSelectParameters()
        {
            // Find and process any ControlParameterExtenders
            foreach (Parameter p in SelectParameters)
            {
                if (p is ControlParameterExtender)
                {
                    ControlParameterExtender qp = (ControlParameterExtender) p;
                    owner.ParameterExtenders.Add(qp.ParameterExtender);
                }
            }

            // Extract values and process them according to ParameterExtender rules, if any
            foreach (DictionaryEntry entry in SelectParameters.GetValues(HttpContext.Current, owner))
            {
                string name = (string) entry.Key;
                ParameterExtender pe = owner.ParameterExtenders[name];
                object value = entry.Value;

                // TODO: delete this
                // commenting this out as it is handled in CreateQueryParameters
                // Exclude empty parameters
                //if (value == null && 
                //    (
                //        owner.ExcludeEmptySelectParameters || 
                //        (pe != null && pe.ExcludeIfEmpty)
                //    )) 
                //    continue;
                
                string stringValue = value as string;
                if (stringValue != null)
                {                    
                    if (pe != null)
                    {
                        stringValue = pe.Prefix + stringValue + pe.Suffix;
                    }
                    if (owner.WildCard != null)
                        stringValue = stringValue.Replace(owner.WildCard, "%");
                    value = stringValue;
                }

                _innerParameters[name] = value;
            }
        }

        /// <summary>
        /// Pre-process the parameter values 
        /// If type information was provided, use it
        /// 
        /// If a null value is used, then NHibernate 
        /// needs a type, so guess it from the DataObjectType
        /// 
        /// </summary>        
        private NamedQueryParameterCollection CreateQueryParameters()
        {
            NamedQueryParameterCollection results = new NamedQueryParameterCollection();
            // set parameters
            foreach (KeyValuePair<string,object> entry in _innerParameters)
            {
                string name = entry.Key;
                ParameterExtender pe = owner.ParameterExtenders[name];
                object value = entry.Value;

                if (owner.ExcludeEmptySelectParameters && IsEmpty(value))
                    continue;

                // The IType of this parameter
                IType type = null;

                if (pe != null)
                {
                    if (pe.Type != null)
                        type = 
                            NHibernateUtil.Entity(
                                System.Web.Compilation.BuildManager.GetType(pe.Type, true, false)
                                );
                    else if (pe.TypeCode != TypeCode.Empty)
                        type = NHibernateUtils.GetType(pe.TypeCode);
                }

                if (value == null && type==null)
                {
                    // try to guess the type from the property name
                    // since NHibernate freaks out with nulls and no type specified
                    if (type == null)
                    {
                        type = NHibernateUtils.GetPropertyType(
                            Session.SessionFactory, 
                            owner.DataObjectType, 
                            name);
                    }
                }

                results.Add(new QueryParameter(name,value,type));
            }
            return results;
        }
        
        /// <summary>
        /// Given the keys from the select args, get the key that
        /// NHibernate uses to identify the entity.
        /// 
        /// This assumes one key is in the keys dictionary
        /// </summary>
        protected virtual object GetKey(IDictionary keys)
        {
            foreach (object value in keys.Values)
            {
                return TypeUtils.ChangeType(value, owner.IdentifierType);
            }
            return null;
			
        }

        /// <summary>
        /// Null-safe way to get metadata
        /// </summary>
        protected virtual IClassMetadata GetMetadata(Type type)
        {
            // Given how data binding works, is there a way to do this?
            if (
                !type.IsPrimitive && 
                type!=typeof(string) &&
                Session.SessionFactory.GetAllClassMetadata().Contains(type)
                )
            {
                // if it is a persisted entity...
              return Session.SessionFactory.GetClassMetadata(type);  
            }
            return null;
        }


        /// <summary>
        /// Convert the value to the appropriate type
        /// If the type is a mapped class, try converting the value to its id and loading it
        /// </summary>
        protected virtual object GetConvertedValue(object value, Type type)
        {

            // special handling for null value types
            if (
                type.IsValueType && 
                (value==null || (value is string && string.IsNullOrEmpty((string)value)))
                )
            {
                return Activator.CreateInstance(type);
            }

            IClassMetadata md = GetMetadata(type);
            if (md != null)
            {
                if (value == null)
                    return null;
                Type returnType = md.MappedClass;
                // if value is already valid...
                if (returnType.IsAssignableFrom(value.GetType()))
                    return value;
                Type idType = md.IdentifierType.ReturnedClass;

                // special handling for nulls
                if (value is string && string.IsNullOrEmpty((string)value))
                    return null;

                object id = GetConvertedValue(value, idType);
                return Session.Load(returnType, id);
            }


            return TypeUtils.ChangeType(value, type);
        }

        /// <summary>
        /// If possible convert the property to the given value
        /// Set the property using reflection
        /// </summary>
        private void SetPropertyValue(object entity, string key, object value)
        {

            if (!owner.GenerateDataTable)
            {
                Type type = owner.DataObjectType;
                PropertyInfo property = type.GetProperty(key);
                object convertedValue = GetConvertedValue(value, property.PropertyType);
                if (property.CanWrite)
                {                    
                    property.SetValue(
                        entity,
                        convertedValue,
                        null);
                }
            }
            // TODO - test this!
            else
            {
                ObjectDataTableAdapter adapter = DataTableAdapter;
                adapter.Data = new object[] { entity };

                // see if we need to change the key
                foreach (PropertyColumn pc in owner.DataTableConfig.Columns)
                {
                    if (pc.ColumnName == key)
                    {
                        if (!string.IsNullOrEmpty(pc.UpdatePropertyName))
                        {
                            key = pc.UpdatePropertyName;
                        }
                        break;
                    }
                }

                Common.Data.PropertyColumn column = adapter.GetMappedProperty(key);
                Type propertyType = column != null ? column.Type : null;

                // Ensure value is correct
                // TODO: Is this necessary?  Seems to be duplicate logic
                if (propertyType!=null)
                    value = GetConvertedValue(value, propertyType);

                adapter.SetProperty(entity, key, value);
            }
        }

        private ObjectDataTableAdapter CreateDataTableAdapter()
        {
            ObjectDataTableAdapter adapter = new ObjectDataTableAdapter();
            DataTableAdapterConfig config = owner.DataTableConfig;
            
            // Add any columns
            foreach (PropertyColumn col in config.Columns)
            {
                if (col.Type == null && col.TypeName != null)
                {
                    col.Type = BuildManager.GetType(col.TypeName, true, false);
                }
                adapter.Columns.Add(col);
            }
            adapter.AutoGenerateColumns = config.AutoGenerateColumns;

            // TODO - don't we have enough (or better!) info to generate the columns here?

            if (config.PropertiesToIgnore != null)
                adapter.PropertiesToIgnore.AddRange(config.PropertiesToIgnore);

            adapter.Type = owner.DataObjectType;
            
            return adapter;
        }

        #endregion

        #region DataSourceView Methods

        /// <summary>
        /// Generate the "Order By" clause
        /// If query is not defined, generate it
        ///     "select from DataObjectType"
        ///     If <see cref="NHibernateDataSource.AutoGenerateWhereClauseFromSelectParams"/>, 
        ///         generate additional WhereFragments from parameter information
        ///     Filter out where fragments whose dependencies are not met
        /// Append the "Order By" expression
        /// Create an IQuery
        /// Set paging parameters, if specified
        /// Execute the query
        /// Issue another query to get total rowcount, if necessary
        ///     <see cref="NHIbernateDataSource.CanRetrieveTotalRowCount"/>
        /// Generate a DataTable, if necessary
        ///     <see cref="NHibernateDataSource.GenerateDataTable"/>
        
        /// </summary>        
        protected override IEnumerable ExecuteSelect(DataSourceSelectArguments arguments)
        {
            // Make sure we can handle the request
            arguments.RaiseUnsupportedCapabilitiesError(this);

            // was an HQL Query defined?
            bool isQueryDefined = this.owner.Query != null;

            ProcessSelectParameters();

            # region Generate Order By Clause

            string orderBy = null;

            string sortExpression = arguments.SortExpression;
            if (string.IsNullOrEmpty(sortExpression))
                sortExpression = owner.DefaultSortExpression;

            if (!string.IsNullOrEmpty(sortExpression))
            {
                if (isQueryDefined)
                {
                    orderBy = string.Format(" order by {0}", sortExpression);
                }
                else
                {
                    string[] parts = sortExpression.Split(new char[] { ' ' }, 2);
                    orderBy = string.Format(" order by o.{0} {1}", parts[0], parts.Length > 1 ? parts[1] : "ASC");
                }
            }

            #endregion

            #region Generate HQL

            string hql = null;
            
            // SELECT clause
            if (isQueryDefined)
            {
                hql = this.owner.Query;
            }
            else
            {                
                hql = 
                hql = string.Format(
                    "select o from {0} o",
                    owner.DataObjectType.Name                    
                    );
            }

            // generate WHERE clause if there is not one
            if (hql.IndexOf(" where ", StringComparison.InvariantCultureIgnoreCase) < 0)
            {
                List<QueryFragment> wheres = new List<QueryFragment>(owner.WhereFragments);
                if (owner.AutoGenerateWhereClauseFromSelectParams)
                    wheres.AddRange(GenerateWhereFragmentsFromSelectParameters());

                FilterUnfulfilledDependencies(wheres);

                string whereClause = BuildWhereClause(wheres);                

                hql = hql + whereClause;
            }


            // ORDER BY
            hql = hql + orderBy;

            #endregion

            // Set up IQuery parameters
            NamedQueryParameterCollection parameters = CreateQueryParameters();


            NHibernateSelectingArgs args1 =
                new NHibernateSelectingArgs(hql,parameters,arguments,false);
            owner.OnSelecting(args1);
            if (args1.Cancel)
            {
                return null;
            }
            hql = args1.Query;

            // Create Query and Set Parameters
            IQuery query = Session.CreateQuery(hql);

            // paging
            if (CanPage)
            {
                if (arguments.MaximumRows > 0)
                    query.SetMaxResults(arguments.MaximumRows);
                if (arguments.StartRowIndex > 0)
                    query.SetFirstResult(arguments.StartRowIndex);
            }

            parameters.Apply(query);



            IList results = query.List();

            #region Set TotalRowCount

            if (arguments.RetrieveTotalRowCount)
            {
                if (arguments.MaximumRows > 0 || arguments.StartRowIndex > 0)
                {
                    if (owner.AutoGenerateCountQuery)
                    {
                        string fromClause = hql.Substring(hql.IndexOf("from ", StringComparison.InvariantCultureIgnoreCase));
                        int orderByIndex = fromClause.IndexOf(" order by ", StringComparison.InvariantCultureIgnoreCase);
                        if (orderByIndex >= 0)
                        {
                            fromClause = fromClause.Substring(0, orderByIndex);
                        }

                        string countHql = "select Count(*) " + fromClause;

                        NHibernateSelectingArgs countArgs =
                            new NHibernateSelectingArgs(countHql, parameters, arguments, true);
                        countHql = countArgs.Query;
                        IQuery countQuery = Session.CreateQuery(countHql);
                        


                        owner.OnSelecting(countArgs);
                        if (countArgs.Cancel)
                        {
                            arguments.TotalRowCount = -1;
                        }
                        else
                        {
                            parameters.Apply(countQuery);
                            arguments.TotalRowCount = (int)countQuery.List()[0];
                        }
                    }
                }
                else
                {
                    arguments.TotalRowCount = results.Count;
                }

            }

            #endregion

            // Generate DataTable, if necessary
            if (owner.GenerateDataTable)
            {
                ObjectDataTableAdapter adapter = DataTableAdapter;
                adapter.Data = results;
                _resultDataTable = adapter.GenerateDataTable();
                return _resultDataTable.DefaultView;
            }
            _resultList = results;

            owner.OnSelected(EventArgs.Empty);

            return _resultList;
        }

        /// <summary>
        /// Call Session.Load to get the object specified by keys
        ///     Currently this requires <see cref="NHibernateDataSource.DataObjectType"/>
        ///         to be populated
        /// Set and changed properties using reflection
        ///     This is interesting in that only changed properties will be modified
        ///         Which means that concurrent modifications will be tolerated 
        ///             unless there is a conflic
        ///     TODO: make this behavior configurable
        /// Update(entity), committing the transaction
        /// unless <see cref="NHibernateDataSource.AutoCommit"/> is false.             
        /// </summary>
        protected override int ExecuteUpdate(IDictionary keys, IDictionary values, IDictionary oldValues)
        {

            object entity = LoadEntity(keys);

            bool changed = false;
            
            foreach (string key in values.Keys)
            {
                if (key.Equals(this.owner.IdentifierProperty)) continue;

                // If property hasn't changed, skip it
                if (values[key] != null && oldValues[key] != null && values[key].Equals(oldValues[key])) 
                    continue;

                SetPropertyValue(entity, key, values[key]);

                changed = true;

            }

            if (changed)
            {
                NHibernateEntityCancelEventArgs args1 = new NHibernateEntityCancelEventArgs(entity);
                owner.OnUpdating(args1);
                if (args1.Cancel)
                {
                    return 0;
                }
                Session.Update(entity);
                if (owner.AutoCommit)
                    CommitAndHandleRollback();

                owner.OnUpdated(EventArgs.Empty);

                this.OnDataSourceViewChanged(EventArgs.Empty);
            }
            return (changed ? 1 : 0);
        }


        protected virtual object LoadEntity(IDictionary keys)
        {
            Type type = owner.DataObjectType;
            object keyValue = GetKey(keys);

            NHibernateEntityLoadingArgs args = new NHibernateEntityLoadingArgs(keys);
            owner.OnLoadingEntity(args);
            if (args.Entity != null)
                return args.Entity;

            return Session.Load(type, keyValue);            
        }

        /// <summary>
        /// Commits the current transaction, starting one if it is not active
        /// On Exception, rolls back and closes the session.
        /// </summary>
        public virtual void CommitAndHandleRollback()
        {
            try
            {
                if (!Session.Transaction.IsActive)
                    Session.BeginTransaction();
                Session.Transaction.Commit();
            }
            catch (Exception)
            {
                Session.Transaction.Rollback();
                Session.Close();
                throw;
            }
        }

        /// <summary>
        /// Load the object and then delete it, committing the transaction
        /// unless <see cref="NHibernateDataSource.AutoCommit"/> is false.
        /// </summary>
        protected override int ExecuteDelete(IDictionary keys, IDictionary oldValues)
        {
            object entity = LoadEntity(keys);

            NHibernateEntityCancelEventArgs args1 =
                new NHibernateEntityCancelEventArgs(entity);
            owner.OnDeleting(args1);
            if (args1.Cancel)
            {
                return 0;
            }

            //TODO: add transaction handling
            //TODO: should it really be necessary to load the object?
            Session.Delete(entity);
            if (owner.AutoCommit)
                CommitAndHandleRollback();

            owner.OnDeleted(EventArgs.Empty);
            this.OnDataSourceViewChanged(EventArgs.Empty);
            return 1;
        }

        /// <summary>
        /// Create a new instance of the DataObjectType
        ///     <see cref="NHibernateDataSource.CreateNewInstance()"/>
        /// Set the properties 
        ///     see <see cref="SetPropertyValue"/>
        /// Save and commit (if <see cref="NHibernateDataSource.AutoCommit"/>
        /// is set to true.
        /// </summary>
        protected override int ExecuteInsert(IDictionary values)
        {

            object entity = owner.CreateNewInstance();
            Type type = entity.GetType();

            foreach (string key in values.Keys)
            {                
                SetPropertyValue(entity, key, values[key]);
            }

            NHibernateEntityCancelEventArgs args1 = 
                new NHibernateEntityCancelEventArgs(entity);
            owner.OnInserting(args1);
            if (args1.Cancel)
            {
                return 0;
            }
			
            Session.Save(entity);
            if (owner.AutoCommit) 
                CommitAndHandleRollback();

            owner.OnInserted(new NHibernateEntityEventArgs(entity));
            this.OnDataSourceViewChanged(EventArgs.Empty);
            return 1;
        }

        #endregion

    }

}
