using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;
using log4net;
using Common.Data.Hibernate;
using Northwind;
using Northwind.Model;
using NHibernate;
using NHE = NHibernate.Expression;
using SC = System.Collections;
using Common.Data;
using System.Data;
using NHibernate.Engine;
using NHibernate.Hql;
using NHibernate.Util;
using HC = NHibernate.Cfg;
using NHibernate.Type;

namespace Northwind.Tests
{
    [TestFixture]
    public class NHibernateTests
    {
        protected ILog Log = LogManager.GetLogger("test");

        [TestFixtureSetUp]
        public void Setup()
        {
            SessionHolder.Interceptor = new NorthwindInterceptor();
        }

        [Test]
        public void TestCustomer()
        {
            Customer cust = (Customer)SessionHolder.Current.Load(typeof(Customer), "ALFKI");
            Log.Debug("cust: " + cust.CompanyName);

            foreach (CustomerDemographic cd in cust.CustomerDemographics)
            {
                Log.Debug(cd.CustomerDesc);
            }
        }


        [Test]
        public void TestOrder()
        {            
            Order o = (Order)SessionHolder.Current.Load(typeof(Order), 10248);
            Log.Debug("order: " + o.OrderDate);
            foreach (OrderDetail d in o.OrderDetails)
            {
                Log.Debug("detail: " +
                    d.Product.ProductName +
                    "(" + d.Product.Category.CategoryName + 
                    " via Supplier: " + d.Product.Supplier.CompanyName +
                    "): "
                    + d.UnitPrice);
            }
            Log.Debug("ship via: " + o.ShipVia.CompanyName);
        }

        [Test]
        public void TestEmployee()
        {
            Employee e = (Employee)SessionHolder.Current.Load(typeof(Employee), 1);
            Log.Debug("employee: " + e.LastName);

            foreach(Territory t in e.Territories) {
                Log.Debug("territory: " + t.TerritoryDescription + " in region: " + t.Region.RegionDescription);
            }
        }

        [Test]
        public void TestProductQuery()
        {
            Log.Debug(
                "Units in stock: " + 
                SessionHolder.Current.CreateQuery(
                    "select p from Product p where p.ProductName = 'chai'")
                    .List<Product>()[0].UnitsInStock
                    );
        }

        [Test]
        public void DeleteRecord(){
            
            ICriteria q = SessionHolder.Current.CreateCriteria(typeof(Customer));
            q.Add(NHE.Expression.Eq("CompanyName", "John Johnson"));

            ITransaction tx = SessionHolder.Current.BeginTransaction();
            try
            {
                Log.Debug("deleting demographic...");
                SessionHolder.Current.Delete("from CustomerDemographic c where c.Id = 'NEWCS'");

                foreach (Customer c in q.List())
                {
                    Log.Debug("deleting customer...");
                    // Delete the orders first
                    // NHibernate should cascade this to delete the OrderDetails 
                    SessionHolder.Current.Delete(
                        "from Northwind.Model.Order o where o.Customer = ?",
                        c, NHibernate.NHibernateUtil.Entity(c.GetType())
                    );
                    SessionHolder.Current.Delete(c);
                }
                tx.Commit();
            }
            catch (Exception)
            {
                tx.Rollback();
            }
        }

        public void DeleteCustomerUsingCascade()
        {
            Log.Debug("deleting customer...");
            ICriteria q = SessionHolder.Current.CreateCriteria(typeof(Customer));
            q.Add(NHE.Expression.Eq("CompanyName", "John Johnson"));

            ITransaction tx = SessionHolder.Current.BeginTransaction();
            try
            {
                foreach (Customer c in q.List())
                {
                    SessionHolder.Current.Delete(c);
                }
                tx.Commit();
            }
            catch (Exception)
            {
                tx.Rollback();
            }
            
        }

        [Test]
        public void TestQueryToDataTable()
        {
            SC.IList list = SessionHolder.Current.CreateQuery("select p from Product p join p.Category").List();
            
            // make sure that the adapter can handle nulls
            SC.ArrayList newList = new SC.ArrayList(list);
            newList.Add(new Product());
            list = newList;

            ObjectDataTableAdapter adapter = new ObjectDataTableAdapter(list);
            adapter.Columns.Add(new PropertyColumn("Category.CategoryName", "CategoryName"));
            adapter.PropertiesToIgnore.Add("IsPersisted");
            DataTable dt = adapter.GenerateDataTable();

            // attempt a modification
            string newName = "New Category Name!";
            dt.Rows[0][dt.Columns.Count-1] = newName;

            Product p = (Product) list[0];
            NUnit.Framework.Assert.AreEqual(newName, p.Category.CategoryName);

            Log.Debug(p.Category.CategoryName);

            Log.Debug("results: " + Environment.NewLine + ToString(dt));
        }

        [Test]
        public void TestSingleColumnQueryToDataTable()
        {
            SC.IList list = SessionHolder.Current.CreateQuery("select p.ProductName from Product p").List();
            list.Insert(0, null);
            ObjectDataTableAdapter adapter = new ObjectDataTableAdapter(list);            

            DataTable dt = adapter.GenerateDataTable();

            NUnit.Framework.Assert.AreEqual(dt.Rows[0][0], System.DBNull.Value);

            string newName = "New Product Name"; ;
            dt.Rows[0][0] = newName;
            NUnit.Framework.Assert.AreEqual(newName, list[0]);

            Log.Debug("results: " + Environment.NewLine + ToString(dt));
        }

        [Test]
        public void TestMultiColumnQueryToDataTable()
        {
            IQuery q = SessionHolder.Current.CreateQuery("select p.Id, p.ProductName, p.Category.CategoryName from Product p");
            SC.IList list = q.List();            
            
            ObjectDataTableAdapter adapter = new ObjectDataTableAdapter(list);

            adapter.AutoGenerateColumns = false;

            int i = 0;
            foreach (IType t in q.ReturnTypes)
            {
                adapter.Columns.Add(new PropertyColumn("[" + i + "]", "GeneratedColumn" + i,t.ReturnedClass));
                i++;
            }

            DataTable dt = adapter.GenerateDataTable();
            Log.Debug("results: " + Environment.NewLine + ToString(dt));


        }

        [Test]
        public void AnotherMultiColumnQueryToDataTable()
        {
            IQuery q = SessionHolder.Current.CreateQuery("select p, p.Category from Product p");
            SC.IList list = q.List();

            ObjectDataTableAdapter adapter = new ObjectDataTableAdapter(list);
            adapter.AutoGenerateColumns = false;
            adapter.Columns.Add(new PropertyColumn("[0].UnitPrice", "UnitPrice"));
            adapter.Columns.Add(new PropertyColumn("[1].Description", "CategoryDescription"));

            DataTable dt = adapter.GenerateDataTable();

            // make sure price is correct type
            Decimal unitPrice = (Decimal) dt.Rows[0][0];
            Log.Debug("1st unit price: " + unitPrice.ToString("C"));

            Log.Debug("results: " + Environment.NewLine + ToString(dt));
        }

        public string ToString(DataTable dt)
        {
            StringBuilder sb = new StringBuilder();
            foreach (DataColumn c in dt.Columns)
            {
                sb.Append(c.ColumnName + ": " + c.DataType + Environment.NewLine);
            }
            foreach (DataColumn c in dt.Columns)
            {
                sb.Append(c + "|");
            }
            sb.Append(Environment.NewLine);

            foreach (DataRow row in dt.Rows)
            {
                foreach (object o in row.ItemArray)
                {
                    sb.Append(
                        (o == null  || o == System.DBNull.Value ? "NULL" : o)
                        + "|");
                }
                sb.Append(Environment.NewLine);
            }
            sb.Append(Environment.NewLine);
            sb.Append("TOTAL: " + dt.Rows.Count);
            return sb.ToString();
        }

        [Test]
        public void TestInsertRecord()
        {


            //just in case
            DeleteRecord();

            Log.Debug("creating customer...");
            Customer customer = new Customer();
            customer.CompanyName = "John Johnson";
            customer.City = "Dallas";
            customer.Id = "NEWCS";
            customer.Orders = new List<Order>();
            customer.CustomerDemographics = new List<CustomerDemographic>();

            CustomerDemographic dem = new CustomerDemographic();
            dem.Id = "NEWCS";            
            dem.CustomerDesc = "John is in a category all by himself";
            dem.Customers = new List<Customer>();
            //SessionHolder.Template.Save(item);
            // we have to explicitly save this since it has an assigned ID
            // technically this should be in the same transaction as below, 
            // but i'm lazy right now
            SessionHolder.Template.Save(dem);  
            
            //associate
            dem.Customers.Add(customer);
            customer.CustomerDemographics.Add(dem);
            
            Order order = new Order();
            
            // associate
            order.Customer = customer;
            customer.Orders.Add(order);

            order.OrderDate = DateTime.Now;
            order.OrderDetails = new List<OrderDetail>();

            OrderDetail item = new OrderDetail();
            
            Log.Debug("get product...");
            //Product chai = SessionHolder.Current.CreateQuery(
            //    "select p from Product p where p.ProductName = 'chai'")
            //    .List<Product>()[0];

            Product chai = SessionHolder.Current.Load<Product>(1);

            item.Product = chai;
            item.Quantity = 3;
            
            // associate
            item.Order = order;
            order.OrderDetails.Add(item);
            
            Log.Debug("saving customer...");


            SessionHolder.Template.Save(customer);

                      
        }

        [Test]
        public void TestAutoGenerateCountQuery()
        {
            string hql = "select p.ProductName, p.Category.CategoryName from Product p";
            string fromClause = hql.Substring(hql.IndexOf("from "));
            string countQuery = "select Count(*) " + fromClause;
            Log.Debug("results: " +  SessionHolder.Current.Find(countQuery)[0]);

        }

        [Test, Obsolete]
        public void TestQueryTranslatorUsage()
        {
            string hql = "select p.ProductName, p.Category.CategoryName from Product p";
            ISessionImplementor session = SessionHolder.Current as ISessionImplementor;
            ISessionFactoryImplementor factory = factory = session.Factory as ISessionFactoryImplementor;
            HC.Settings settings = factory.Settings;

            string[] concreteQueryStrings = QuerySplitter.ConcreteQueries(hql, factory);
            int length = concreteQueryStrings.Length;
            IQueryTranslator[] queries = new IQueryTranslator[length];
            SC.IDictionary enabledFilters = CollectionHelper.EmptyMap;
            for (int i = 0; i < length; i++)
            {
                queries[i] = settings.QueryTranslatorFactory.CreateQueryTranslator(
                    concreteQueryStrings[i], enabledFilters, factory);
            }

            foreach (IQueryTranslator q in queries)
            {
                q.Compile(settings.QuerySubstitutions, false);               
                Log.Debug(q.ReturnAliases.Length);
            }
            
        }

    }
}
